@@ -79,7 +79,8 @@ cuda_minor_mismatch_ok = {
          "11.4",
          "11.5",
          "11.6",
-         "11.7"],
+         "11.7",
+         "11.8"],
 }
 
 
@@ -406,18 +407,6 @@ class OpBuilder(ABC):
                 return '-D__AVX256__'
         return '-D__SCALAR__'
 
-    def python_requirements(self):
-        '''
-        Override if op wants to define special dependencies, otherwise will
-        take self.name and load requirements-<op-name>.txt if it exists.
-        '''
-        path = f'requirements/requirements-{self.name}.txt'
-        requirements = []
-        if os.path.isfile(path):
-            with open(path, 'r') as fd:
-                requirements = [r.strip() for r in fd.readlines()]
-        return requirements
-
     def command_exists(self, cmd):
         if '|' in cmd:
             cmds = cmd.split("|")
@@ -487,13 +476,6 @@ class OpBuilder(ABC):
         self.jit_mode = True
         from torch.utils.cpp_extension import load
 
-        # Ensure directory exists to prevent race condition in some cases
-        ext_path = os.path.join(
-            os.environ.get('TORCH_EXTENSIONS_DIR',
-                           DEFAULT_TORCH_EXTENSION_PATH),
-            self.name)
-        os.makedirs(ext_path, exist_ok=True)
-
         start_build = time.time()
         sources = [self.deepspeed_src_path(path) for path in self.sources()]
         extra_include_paths = [
